@@ -94,6 +94,9 @@ import Control.Monad.Catch
     ( MonadCatch(..)
     , MonadThrow(..)
     , MonadMask(..)
+#if MIN_VERSION_exceptions(0,10,0)
+    , ExitCase(..)
+#endif
     , mask_
     , uninterruptibleMask_
     , catchAll
@@ -140,7 +143,6 @@ import qualified Data.Map as M
 import Data.Monoid (Monoid)
 import Pipes (Proxy, Effect, Effect', runEffect)
 import Pipes.Internal (Proxy(..))
-import Pipes.Lift (liftCatchError)
 
 data Restore m = Unmasked | Masked (forall x . m x -> m x)
 
@@ -181,9 +183,27 @@ liftMask maskVariant k = do
     loop $ k unmask
 
 instance (MonadMask m, MonadIO m) => MonadMask (Proxy a' a b' b m) where
-    mask                = liftMask mask
+    mask = liftMask mask
+
     uninterruptibleMask = liftMask uninterruptibleMask
 
+#if MIN_VERSION_exceptions(0,10,0)
+    generalBracket acquire release_ use = mask $ \unmasked -> do
+      a <- acquire
+      let action = do
+              b <- use a
+              return (ExitCaseSuccess b, ExitCaseSuccess_ b)
+      let handler e = return (ExitCaseException e, ExitCaseException_ e)
+      (exitCase, exitCase_) <- unmasked action `catch` handler
+      c <- release_ a exitCase
+      case exitCase_ of
+          ExitCaseException_ e -> throwM e
+          ExitCaseSuccess_ b   -> return (b, c)
+
+-- | This is to avoid an unnecessary partial pattern match in `generalBracket`
+data ExitCase_ a = ExitCaseSuccess_ a | ExitCaseException_ SomeException
+#endif
+
 data Finalizers m = Finalizers
     { _nextKey    :: !Integer
     , _finalizers :: !(M.Map Integer (m ()))
